(**
  To be viewed in Isabelle/jedit IDE. In a normal text editor, this will just be UNREADABLE!

  http://isabelle.in.tum.de

*)
theory Edka_Impl_Demo
  (*
    We import some background theory
  *)
  imports 
    Edka_Impl_Demo_Background
begin


context Network begin

(*
  We add an assertion to the abstract algorithm, stating that
  the selected path is actually an augmenting path.
  
  This is straightforward when proving the abstract algorithm correct.
*)

definition edmonds_karp where "edmonds_karp \<equiv> do {
  let f = (\<lambda>_. 0);
  (f,_) \<leftarrow> while\<^sub>T
    (\<lambda>(f,brk). \<not>brk) 
    (\<lambda>(f,_). do {
      p \<leftarrow> select p. Graph.isShortestPath (cf_of f) s p t;
      case p of 
        None \<Rightarrow> return (f,True)
      | Some p \<Rightarrow> do {
          assert (NPreflow.isAugmentingPath c s t f p); \<comment> \<open>Added assertion\<close>
          assert (Graph.isShortestPath (cf_of f) s p t);
          let f = NFlow.augment_with_path c f p;
          return (f, False)
        }  
    })
    (f,False);
  return f 
}"

(* Actually, the correctness proof does not change at all! *)

abbreviation edka_invar where "edka_invar \<equiv> \<lambda>(f,brk). 
        NFlow c s t f 
      \<and> (brk \<longrightarrow> (\<forall>p. \<not>NPreflow.isAugmentingPath c s t f p))
    "  

definition "edka_wf_rel \<equiv> inv_image 
  (less_than_bool <*lex*> measure (\<lambda>cf. ek_analysis_defs.ekMeasure cf s t))
  (\<lambda>(f,brk). (\<not>brk,residualGraph c f))"

lemma edka_wf_rel_wf[simp, intro!]: "wf edka_wf_rel"
  unfolding edka_wf_rel_def by auto

lemma edmonds_karp_correct: "edmonds_karp \<le> (spec f. isMaxFlow f)"
  unfolding edmonds_karp_def
  apply (refine_vcg WHILET_rule[where I=fofu_invar and R=edka_wf_rel])
  apply (auto simp: 
    zero_flow 
    NFlow.augment_pres_nflow 
    NFlow.augmenting_path_not_empty
    NFlow.noAugPath_iff_maxFlow[symmetric]
    NFlow.augment_with_path_def
    edka_wf_rel_def NFlow.shortest_path_decr_ek_measure
    NFlow.shortest_is_augmenting
    dest: NFlow.augmenting_path_imp_shortest
  )
  done
  
  

(** At this point, we have *)

term cfi_rel
thm cfi_rel_def
  
thm zero_flow_cfi_refine   \<comment> \<open>To refine initialization\<close>
thm augment_cfi_refine     \<comment> \<open>To refine augmentation\<close>
thm cf_of_cfi_rel          \<comment> \<open>Shortest path computed on residual graph\<close>
thm flow_of_cfi_rel        \<comment> \<open>Compute flow from residual graph\<close>
  
  
definition edmonds_karp2 where "edmonds_karp2 \<equiv> do {
  let cf = c;
  assert (RGraph c s t cf);
  (cf,_) \<leftarrow> while\<^sub>T
    (\<lambda>(cf,brk). \<not>brk) 
    (\<lambda>(cf,_). do {
      assert (RGraph c s t cf);
      p \<leftarrow> select p. Graph.isShortestPath cf s p t;
      case p of 
        None \<Rightarrow> return (cf,True)
      | Some p \<Rightarrow> do {
          assert (Graph.isShortestPath cf s p t);
          let cf = Graph.augment_cf cf (set p) (resCap_cf cf p);
          return (cf, False)
        }  
    })
    (cf,False);
  assert (RGraph c s t cf);
  return (flow_of_cf cf)
}"
  

(* Do refinement *)  

text \<open>Automatic proof\<close>

lemma edka2_refine: "edmonds_karp2 \<le> \<Down>Id edmonds_karp"
  unfolding edmonds_karp2_def edmonds_karp_def
  \<comment> \<open>Register \<open>cfi_rel\<close> for relation guessing heuristics\<close>
  supply RELATESI[of cfi_rel, refine_dref_RELATES]
  \<comment> \<open>Invoke VCG\<close>
  apply refine_vcg
  \<comment> \<open>Guess Relations\<close>
  apply refine_dref_type
  \<comment> \<open>Prove VCs\<close>
  apply (auto simp: zero_flow_cfi_refine cf_of_cfi_rel flow_of_cfi_rel 
    augment_cfi_refine cfi_rel_imp_invar)
    subgoal sorry
  done


(* Use BFS, implement bottleneck computation and augmentation *)  
    
definition edmonds_karp3 where "edmonds_karp3 \<equiv> do {
  let cf = c;
  (cf,_) \<leftarrow> while\<^sub>T
    (\<lambda>(cf,brk). \<not>brk) 
    (\<lambda>(cf,_). do {
      assert (RGraph c s t cf);
      p\<leftarrow> Graph.bfs cf s t;
      case p of 
        None \<Rightarrow> return (cf,True)
      | Some p \<Rightarrow> do {
          bn \<leftarrow> resCap_cf_impl cf p;
          cf \<leftarrow> augment_cf_impl cf p bn;
          return (cf, False)
        }  
    })
    (cf,False);
  assert (RGraph c s t cf);
  return (flow_of_cf cf)
}"

(*
  We customize the specification of BFS to our needs
*)
thm Graph.bfs_correct
lemma bfs_spec': 
  "RGraph c s t cf \<Longrightarrow> Graph.bfs cf s t \<le> (select p. Graph.isShortestPath cf s p t)"
  unfolding SELECT_def
  apply (split if_split; refine_vcg Graph.bfs_correct)
  apply (vc_solve split: option.splits)
  subgoal
    using RGraph.this_loc_rpg RPreGraph.resV_netV by blast 
  subgoal
    using Graph.Efin_imp_Vfin RGraph.EX_RG by blast 
  subgoal
    using Graph.connected_def Graph.shortestPath_is_path by blast 
  subgoal
    using RGraph.this_loc_rpg RPreGraph.resV_netV by blast 
  subgoal
    using Graph.Efin_imp_Vfin RGraph.EX_RG by blast 
  done


lemma edka3_refine: "edmonds_karp3 \<le> \<Down>Id edmonds_karp2"
  unfolding edmonds_karp3_def edmonds_karp2_def
  apply (rewrite in "let cf = Graph.augment_cf _ _ _ in _" Let_def)
  apply refine_rcg
  apply refine_dref_type
  apply (vc_solve simp: bfs_spec')
  
  (* We have arranged the proof such that the structural change remains, and
    we can prove it explicitly here. *)
  apply (drule Graph.shortestPath_is_simple)
  apply (frule (1) RGraph.resCap_cf_impl_refine)
  apply (frule (1) RGraph.augment_cf_impl_refine)
  apply (auto simp: pw_le_iff refine_pw_simps)
  done
  
  

  text \<open>We split the algorithm into a tabulation function, and the 
    running of the actual algorithm:\<close>
  definition "edmonds_karp4_tabulate am \<equiv> do {
    cf \<leftarrow> init_cf;
    am \<leftarrow> init_ps am;
    return (cf,am)
  }"

  definition "edmonds_karp4_run cf am \<equiv> do {
    (cf,_) \<leftarrow> while\<^sub>T 
      (\<lambda>(cf,brk). \<not>brk) 
      (\<lambda>(cf,_). do {
        assert (RGraph c s t cf);
        p \<leftarrow> bfs2_op am cf;
        case p of 
          None \<Rightarrow> return (cf,True)
        | Some p \<Rightarrow> do {
            bn \<leftarrow> resCap_cf_impl cf p;
            cf \<leftarrow> augment_cf_impl cf p bn;
            return (cf, False)
          }  
      })
      (cf,False);
    assert (RGraph c s t cf);
    f \<leftarrow> compute_rflow cf;  
    return f
  }"

  definition "edmonds_karp4 am \<equiv> do {
    (cf,am) \<leftarrow> edmonds_karp4_tabulate am;
    edmonds_karp4_run cf am
  }"

  lemma edka5_refine: "\<lbrakk>is_adj_map am\<rbrakk> \<Longrightarrow> edmonds_karp4 am \<le> \<Down>Id edmonds_karp3"
    unfolding edmonds_karp4_def edmonds_karp4_tabulate_def edmonds_karp4_run_def
      edmonds_karp3_def init_cf_def compute_rflow_def
      init_ps_def Let_def nres_monad_laws bfs2_op_def
    apply refine_rcg
    apply refine_dref_type
    apply (vc_solve (nopre) solve: asm_rl)
    apply clarsimp
    apply (rule refine_IdD)
    apply (rule Graph.bfs2_refine)
    apply (simp add: RPreGraph.resV_netV[OF RGraph.this_loc_rpg])
    apply (simp add: RGraph.rg_succ_ref)
    done

end

  subsection \<open>Imperative Implementation\<close>  
  text \<open>In this section we provide an efficient imperative implementation,
    using the Sepref tool. It is mostly technical, setting up the mappings
    from abstract to concrete data structures, and then refining the algorithm,
    function by function.  
    \<close>

  text \<open>
    This is also the point where we have to choose the implementation of 
    capacities. Up to here, they have been a polymorphic type with a
    typeclass constraint of being a linearly ordered integral domain.
    Here, we switch to @{typ [source] capacity_impl} (@{typ capacity_impl}).
    \<close>
  locale Network_Impl = Network c s t for c :: "capacity_impl graph" and s t

  text \<open>Moreover, we assume that the nodes are natural numbers less 
    than some number @{term N}, which will become an additional parameter 
    of our algorithm. \<close>
  locale Edka_Impl = Network_Impl +
    fixes N :: nat
    assumes V_ss: "V\<subseteq>{0..<N}"
  begin  
    lemma this_loc: "Edka_Impl c s t N" by unfold_locales

    lemma E_ss: "E \<subseteq> {0..<N}\<times>{0..<N}" using E_ss_VxV V_ss by auto

    lemma mtx_nonzero_iff[simp]: "mtx_nonzero c = E" unfolding E_def by (auto simp: mtx_nonzero_def)

    lemma mtx_nonzeroN: "mtx_nonzero c \<subseteq> {0..<N}\<times>{0..<N}" using E_ss by simp

    lemma [simp]: "v\<in>V \<Longrightarrow> v<N" using V_ss by auto


    text \<open>Declare some variables to Sepref. \<close>
    lemmas [id_rules] = 
      itypeI[Pure.of N "TYPE(nat)"]  
      itypeI[Pure.of s "TYPE(node)"]  
      itypeI[Pure.of t "TYPE(node)"]  
      itypeI[Pure.of c "TYPE(capacity_impl graph)"]  
    text \<open>Instruct Sepref to not refine these parameters. This is expressed
      by using identity as refinement relation.\<close>
    lemmas [sepref_import_param] = 
      IdI[of N]
      IdI[of s]
      IdI[of t]
      (*IdI[of c]*)

      
    lemma [sepref_fr_rules]: "(uncurry0 (Heap_Monad.return c),uncurry0 (return c))\<in>unit_assn\<^sup>k \<rightarrow>\<^sub>a pure (nat_rel\<times>\<^sub>rnat_rel \<rightarrow> int_rel)"
      apply sepref_to_hoare by sep_auto


    subsubsection \<open>Implementation of Adjacency Map by Array\<close>  
    definition "is_am am psi 
      \<equiv> \<exists>\<^sub>Al. psi \<mapsto>\<^sub>a l 
          * \<up>(length l = N \<and> (\<forall>i<N. l!i = am i) 
              \<and> (\<forall>i\<ge>N. am i = []))"
  
    lemma is_am_precise[safe_constraint_rules]: "precise (is_am)"
      apply rule
      unfolding is_am_def
      apply clarsimp
      apply (rename_tac l l')
      apply prec_extract_eqs
      apply (rule ext)
      apply (rename_tac i)
      apply (case_tac "i<length l'")
      apply fastforce+
      done

    sepref_decl_intf i_ps is "nat \<Rightarrow> nat list" 

    definition (in -) "ps_get_imp psi u \<equiv> Array.nth psi u"

    lemma [def_pat_rules]: "Network.ps_get_op$c \<equiv> UNPROTECT ps_get_op" by simp
    sepref_register "PR_CONST ps_get_op" :: "i_ps \<Rightarrow> node \<Rightarrow> node list nres"

    lemma ps_get_op_refine[sepref_fr_rules]: 
      "(uncurry ps_get_imp, uncurry (PR_CONST ps_get_op)) 
        \<in> is_am\<^sup>k *\<^sub>a (pure Id)\<^sup>k \<rightarrow>\<^sub>a list_assn (pure Id)"
      unfolding list_assn_pure_conv
      apply sepref_to_hoare
      using V_ss
      by (sep_auto 
            simp: is_am_def pure_def ps_get_imp_def 
            simp: ps_get_op_def refine_pw_simps)

    lemma is_pred_succ_no_node: "\<lbrakk>is_adj_map a; u\<notin>V\<rbrakk> \<Longrightarrow> a u = []"
      unfolding is_adj_map_def V_def
      by auto

    lemma [sepref_fr_rules]: "(Array.make N, PR_CONST init_ps) 
      \<in> (pure Id)\<^sup>k \<rightarrow>\<^sub>a is_am" 
      apply sepref_to_hoare
      using V_ss
      by (sep_auto simp: init_ps_def refine_pw_simps is_am_def pure_def
        intro: is_pred_succ_no_node)

    lemma [def_pat_rules]: "Network.init_ps$c \<equiv> UNPROTECT init_ps" by simp
    sepref_register "PR_CONST init_ps" :: "(node \<Rightarrow> node list) \<Rightarrow> i_ps nres"

    subsubsection \<open>Implementation of Capacity Matrix by Array\<close>  
    lemma [def_pat_rules]: "Network.cf_get$c \<equiv> UNPROTECT cf_get" by simp
    lemma [def_pat_rules]: "Network.cf_set$c \<equiv> UNPROTECT cf_set" by simp

    sepref_register 
      "PR_CONST cf_get" :: "capacity_impl i_mtx \<Rightarrow> edge \<Rightarrow> capacity_impl nres"
    sepref_register 
      "PR_CONST cf_set" :: "capacity_impl i_mtx \<Rightarrow> edge \<Rightarrow> capacity_impl 
        \<Rightarrow> capacity_impl i_mtx nres"

    text \<open>We have to link the matrix implementation, which encodes the bound, 
      to the abstract assertion of the bound\<close>

    sepref_definition cf_get_impl is "uncurry (PR_CONST cf_get)" :: "(asmtx_assn N id_assn)\<^sup>k *\<^sub>a (prod_assn id_assn id_assn)\<^sup>k \<rightarrow>\<^sub>a id_assn"
      unfolding PR_CONST_def cf_get_def[abs_def]
      by sepref
    lemmas [sepref_fr_rules] = cf_get_impl.refine
    lemmas [sepref_opt_simps] = cf_get_impl_def

    sepref_definition cf_set_impl is "uncurry2 (PR_CONST cf_set)" :: "(asmtx_assn N id_assn)\<^sup>d *\<^sub>a (prod_assn id_assn id_assn)\<^sup>k *\<^sub>a id_assn\<^sup>k \<rightarrow>\<^sub>a asmtx_assn N id_assn"
      unfolding PR_CONST_def cf_set_def[abs_def]
      by sepref
    lemmas [sepref_fr_rules] = cf_set_impl.refine
    lemmas [sepref_opt_simps] = cf_set_impl_def


    sepref_thm init_cf_impl is "uncurry0 (PR_CONST init_cf)" :: "unit_assn\<^sup>k \<rightarrow>\<^sub>a asmtx_assn N id_assn"
      unfolding PR_CONST_def init_cf_def 
      using E_ss
      apply (rewrite op_mtx_new_def[of c, symmetric])
      apply (rewrite amtx_fold_custom_new[of N N])
      by sepref

    concrete_definition (in -) init_cf_impl uses Edka_Impl.init_cf_impl.refine_raw is "(uncurry0 ?f,_)\<in>_" 
    prepare_code_thms (in -) init_cf_impl_def
    lemmas [sepref_fr_rules] = init_cf_impl.refine[OF this_loc]  

    (* TODO: Use sepref to synthesize the get-operations! *)
    lemma amtx_cnv: "amtx_assn N M id_assn = IICF_Array_Matrix.is_amtx N M" 
      by (simp add: amtx_assn_def)

    lemma [def_pat_rules]: "Network.init_cf$c \<equiv> UNPROTECT init_cf" by simp
    sepref_register "PR_CONST init_cf" :: "capacity_impl i_mtx nres"

    subsubsection \<open>Representing Result Flow as Residual Graph\<close>
    definition (in Network_Impl) "is_rflow N f cfi 
      \<equiv> \<exists>\<^sub>Acf. asmtx_assn N id_assn cf cfi * \<up>(RGraph c s t cf \<and> f = flow_of_cf cf)"
    lemma is_rflow_precise[safe_constraint_rules]: "precise (is_rflow N)"
      apply rule
      unfolding is_rflow_def
      apply (clarsimp simp: amtx_assn_def)
      apply prec_extract_eqs
      apply simp
      done

    sepref_decl_intf i_rflow is "nat\<times>nat \<Rightarrow> int"

    lemma [sepref_fr_rules]: 
      "(\<lambda>cfi. Heap_Monad.return cfi, PR_CONST compute_rflow) \<in> (asmtx_assn N id_assn)\<^sup>d \<rightarrow>\<^sub>a is_rflow N"
      unfolding amtx_cnv
      apply sepref_to_hoare
      apply (sep_auto simp: amtx_cnv compute_rflow_def is_rflow_def refine_pw_simps hn_ctxt_def)
      done

    lemma [def_pat_rules]: 
      "Network.compute_rflow$c$s$t \<equiv> UNPROTECT compute_rflow" by simp
    sepref_register 
      "PR_CONST compute_rflow" :: "capacity_impl i_mtx \<Rightarrow> i_rflow nres"


    subsubsection \<open>Implementation of Functions\<close>  

    declare intf_of_assnI[of is_am, where 'a=i_ps, intf_of_assn]
    lemmas [sepref_fr_rules] = HOL_list_empty_hnr
    
    
    sepref_thm rg_succ2_impl is "uncurry2 (PR_CONST rg_succ2)" 
      :: "is_am\<^sup>k *\<^sub>a (asmtx_assn N id_assn)\<^sup>k *\<^sub>a nat_assn\<^sup>k \<rightarrow>\<^sub>a list_assn nat_assn"
      unfolding rg_succ2_def APP_def monadic_filter_rev_def monadic_filter_rev_aux_def PR_CONST_def
      (* TODO: Make setting up combinators for sepref simpler, then we do not need to unfold! *)
      using [[id_debug, goals_limit = 1]]
      by sepref_dbg_keep
    concrete_definition (in -) succ_imp uses Edka_Impl.rg_succ2_impl.refine_raw is "(uncurry2 ?f,_)\<in>_"
    prepare_code_thms (in -) succ_imp_def
    lemmas succ_imp_refine[sepref_fr_rules] = succ_imp.refine[OF this_loc]
    sepref_register 
      "rg_succ2" :: "i_ps \<Rightarrow> capacity_impl i_mtx \<Rightarrow> node \<Rightarrow> node list nres"

    (* TODO: Move *)      
    lemma [sepref_import_param]: "(min,min)\<in>Id\<rightarrow>Id\<rightarrow>Id" by simp


    abbreviation "is_path \<equiv> list_assn (prod_assn (pure Id) (pure Id))"

    
    sepref_thm resCap_imp_impl is "uncurry (PR_CONST resCap_cf_impl)" 
      :: "(asmtx_assn N id_assn)\<^sup>k *\<^sub>a is_path\<^sup>k \<rightarrow>\<^sub>a int_assn"
      unfolding resCap_cf_impl_def APP_def PR_CONST_def
      by sepref
    concrete_definition (in -) resCap_imp uses Edka_Impl.resCap_imp_impl.refine_raw is "(uncurry ?f,_)\<in>_"
    prepare_code_thms (in -) resCap_imp_def
    lemmas resCap_impl_refine[sepref_fr_rules] = resCap_imp.refine[OF this_loc]
    sepref_register "resCap_cf_impl" 
      :: "capacity_impl i_mtx \<Rightarrow> path \<Rightarrow> capacity_impl nres"
   
    sepref_thm augment_imp is "uncurry2 (PR_CONST augment_cf_impl)" :: "((asmtx_assn N id_assn)\<^sup>d *\<^sub>a (is_path)\<^sup>k *\<^sub>a (pure Id)\<^sup>k \<rightarrow>\<^sub>a asmtx_assn N id_assn)"
      unfolding augment_cf_impl_def[abs_def] augment_edge_impl_def PR_CONST_def
      using [[id_debug, goals_limit = 1]]
      by sepref 
    concrete_definition (in -) augment_imp uses Edka_Impl.augment_imp.refine_raw is "(uncurry2 ?f,_)\<in>_"
    prepare_code_thms (in -) augment_imp_def
    lemmas augment_impl_refine[sepref_fr_rules] = augment_imp.refine[OF this_loc]
    sepref_register "augment_cf_impl" 
      :: "capacity_impl i_mtx \<Rightarrow> path \<Rightarrow> capacity_impl \<Rightarrow> capacity_impl i_mtx nres"

    sublocale bfs: Impl_Succ 
      "snd" 
      "TYPE(i_ps \<times> capacity_impl i_mtx)" 
      "PR_CONST (\<lambda>(am,cf). rg_succ2 am cf)" 
      "prod_assn is_am (asmtx_assn N id_assn)" 
      "\<lambda>(am,cf). succ_imp N am cf"
      unfolding APP_def
      apply unfold_locales
      apply (simp add: fold_partial_uncurry)
      apply (rule hfref_cons[OF succ_imp_refine[unfolded PR_CONST_def]])
      by auto
      
    definition (in -) "bfsi' N s t psi cfi 
      \<equiv> bfs_impl (\<lambda>(am, cf). succ_imp N am cf) (psi,cfi) s t"

    lemma [sepref_fr_rules]: 
      "(uncurry (bfsi' N s t),uncurry (PR_CONST bfs2_op)) 
        \<in> is_am\<^sup>k *\<^sub>a (asmtx_assn N id_assn)\<^sup>k \<rightarrow>\<^sub>a option_assn is_path"
      unfolding bfsi'_def[abs_def] bfs2_op_def[abs_def] 
      using bfs.bfs_impl_fr_rule unfolding bfs.op_bfs_def[abs_def]
      apply (clarsimp simp: hfref_def all_to_meta)
      apply (rule hn_refine_cons[rotated])
      apply rprems
      apply (sep_auto simp: pure_def intro!: enttI)
      apply (sep_auto simp: pure_def)
      apply (sep_auto simp: pure_def)
      done
    sepref_register "bfs2_op" 
      :: "i_ps \<Rightarrow> capacity_impl i_mtx \<Rightarrow> path option nres"  

      
    sepref_thm edka_imp_tabulate_impl 
      is "PR_CONST edmonds_karp4_tabulate" :: "id_assn\<^sup>k \<rightarrow>\<^sub>a asmtx_assn N int_assn \<times>\<^sub>a is_am"
      unfolding edmonds_karp4_tabulate_def PR_CONST_def
      by sepref
    concrete_definition (in -) edka_imp_tabulate 
      uses Edka_Impl.edka_imp_tabulate_impl.refine_raw is "(?f,_):_"
    prepare_code_thms (in -) edka_imp_tabulate_def
    lemmas edka_imp_tabulate_refine[sepref_fr_rules] = edka_imp_tabulate.refine[OF this_loc]
    sepref_register "edmonds_karp4_tabulate"
      :: "(node \<Rightarrow> node list) \<Rightarrow> (capacity_impl i_mtx \<times> i_ps) nres"


    sepref_thm edka_imp_run_impl is "uncurry (PR_CONST edmonds_karp4_run)" :: "(asmtx_assn N id_assn)\<^sup>d*\<^sub>ais_am\<^sup>k \<rightarrow>\<^sub>a is_rflow N"
      unfolding edmonds_karp4_run_def PR_CONST_def
      by sepref
    concrete_definition (in -) edka_imp_run uses Edka_Impl.edka_imp_run_impl.refine_raw is "(uncurry ?f,_)\<in>_"
    prepare_code_thms (in -) edka_imp_run_def
    lemmas edka_imp_run_refine[sepref_fr_rules] = edka_imp_run.refine[OF this_loc]
    sepref_register "edmonds_karp4_run" 
      :: "capacity_impl i_mtx \<Rightarrow> i_ps \<Rightarrow> i_rflow nres"


    sepref_thm edka_imp_impl is "PR_CONST edmonds_karp4" :: "id_assn\<^sup>k \<rightarrow>\<^sub>a is_rflow N"
      unfolding edmonds_karp4_def PR_CONST_def
      by sepref
    concrete_definition (in -) edka_imp uses Edka_Impl.edka_imp_impl.refine_raw
    prepare_code_thms (in -) edka_imp_def
    lemmas edka_imp_refine = edka_imp.refine[OF this_loc]

  end

  export_code edka_imp checking SML_imp

  subsection \<open>Correctness Theorem for Implementation\<close>
  text \<open>We combine all refinement steps to derive a correctness 
    theorem for the implementation\<close>
  context Network_Impl begin
    theorem edka_imp_correct: 
      assumes VN: "Graph.V c \<subseteq> {0..<N}"
      assumes ABS_PS: "is_adj_map am"
      shows "
        <emp> 
          edka_imp c s t N am 
        <\<lambda>fi. \<exists>\<^sub>Af. is_rflow N f fi * \<up>(isMaxFlow f)>\<^sub>t"
    proof -
      interpret Edka_Impl by unfold_locales fact

      note edka5_refine[OF ABS_PS]
      also note edka3_refine    
      also note edka2_refine 
      also note edmonds_karp_correct
      finally have "edmonds_karp4 am \<le> SPEC isMaxFlow" .
      from hn_refine_ref[OF this edka_imp_refine[to_hnr, unfolded autoref_tag_defs], of am]
      show ?thesis by (simp add: hn_refine_def)
    qed
  end    


  (*
    In the AFP-Entry:
      Removing preconditions of correctness thm, 
      by loading network from list, and implementing checker that
      checks for valid flow network.

      CLI for Fofu-algorithm
      
      Also Push-Relabel algorithm
    
  *)




end
