theory Edka_Impl_Demo_Background
  imports 
    EdmondsKarp_Maxflow.FordFulkerson_Algo 
    EdmondsKarp_Maxflow.EdmondsKarp_Termination_Abstract
    EdmondsKarp_Maxflow.Augmenting_Path_BFS

begin

(* Shut up some confusing warnings \<dots> *)
no_notation Heap_Monad.return ("return")

(* TODO: Move *)
  lemma pure_id_refl[simp]: "pure Id a a = emp"
    by (auto simp: pure_def)

  lemma val_id_refl[simp]: "hn_val Id a a = emp"  
    by (auto simp: pure_def hn_ctxt_def)


context Network begin

  (******** Some background theory on flows **)
  text \<open>We define the relation between residual graphs and flows\<close>
  definition "cfi_rel \<equiv> br flow_of_cf (RGraph c s t)"
    
  lemma cfi_rel_imp_invar: "(cf,f)\<in>cfi_rel \<Longrightarrow> RGraph c s t cf"
    by (simp add: cfi_rel_def in_br_conv)

  
  text \<open>\<^const>\<open>cfi_rel\<close> can also be characterized the other way round, i.e., 
    mapping flows to residual graphs:\<close>
  lemma cfi_rel_alt: "cfi_rel = {(cf,f). cf = residualGraph c f \<and> NFlow c s t f}"
    unfolding cfi_rel_def br_def
    by (auto 
        simp: NFlow.is_RGraph RGraph.is_NFlow 
        simp: RPreGraph.rg_fo_inv[OF RGraph.this_loc_rpg]
        simp: NPreflow.fo_rg_inv[OF NFlow.axioms(1)])
    
  lemma cf_of_cfi_rel: "(cf,f)\<in>cfi_rel \<Longrightarrow> cf_of f = cf"
    unfolding cfi_rel_alt
    by (auto simp: in_br_conv)
  
  lemma flow_of_cfi_rel: "(cf,f)\<in>cfi_rel \<Longrightarrow> flow_of_cf cf = f"
    unfolding cfi_rel_def
    by (auto simp: in_br_conv)
    
  
  
  lemma residualGraph_zero_flow: "residualGraph c (\<lambda>_. 0) = c" 
    unfolding residualGraph_def by (auto intro!: ext)
    
  lemma flow_of_c: "flow_of_cf c = (\<lambda>_. 0)"
    by (auto simp add: flow_of_cf_def[abs_def])
  
  lemma zero_flow_cfi_refine: "(c,\<lambda>_. 0) \<in> cfi_rel"  
  proof -
    have "NFlow c s t (\<lambda>_. 0)" by (rule zero_flow)
    hence "RGraph c s t c" 
      by (auto dest: NFlow.is_RGraph simp: residualGraph_zero_flow) 
    then show ?thesis
      by (auto simp: cfi_rel_def in_br_conv flow_of_c)
  qed  
  
  
  definition "resCap_cf cf p \<equiv> Min {cf e | e. e\<in>set p}"
  lemma (in NFlow) resCap_cf_refine: "resCap_cf cf p = resCap p"
    unfolding resCap_cf_def resCap_def ..
    
  lemma (in NFlow) augment_cfi_refine_aux: (* For snippet *)
    assumes AUG: "isAugmentingPath p"
    shows "residualGraph c (augment (augmentingFlow p)) (u,v) = (
      if (u,v)\<in>set p then (residualGraph c f (u,v) - resCap p)
      else if (v,u)\<in>set p then (residualGraph c f (u,v) + resCap p)
      else residualGraph c f (u,v))"
    using augment_alt[OF AUG] by (auto simp: Graph.augment_cf_def)

        
        
  lemma augment_cfi_refine:
    assumes R: "(cf,f)\<in>cfi_rel"
    assumes AUG: "NPreflow.isAugmentingPath c s t f p"
    shows "(Graph.augment_cf cf (set p) (resCap_cf cf p), 
        NFlow.augment_with_path c f p) \<in> cfi_rel"
  proof -    
    from R have [simp]: "cf = residualGraph c f" and "NFlow c s t f"
      by (auto simp: cfi_rel_alt br_def)
    then interpret f: NFlow c s t f by simp
    
    show ?thesis 
      unfolding f.augment_with_path_def
    proof (simp add: cfi_rel_alt; safe del: ext intro!: ext)
      fix u v
      show "Graph.augment_cf f.cf (set p) (resCap_cf f.cf p) (u,v) 
            = residualGraph c (f.augment (f.augmentingFlow p)) (u,v)"
        unfolding f.augment_cfi_refine_aux[OF AUG]
        unfolding f.cf.augment_cf_def
        by (auto simp: f.resCap_cf_refine)
    qed (rule f.augment_pres_nflow[OF AUG])
  qed  
  
(**********)
 

end

  context Network begin


    subsection \<open>Implementation of Bottleneck Computation and Augmentation\<close>  
    text \<open>We will access the capacities in the residual graph
      only by a get-operation, which asserts that the edges are valid\<close>
    
    abbreviation (input) valid_edge :: "edge \<Rightarrow> bool" where
      "valid_edge \<equiv> \<lambda>(u,v). u\<in>V \<and> v\<in>V"

    definition cf_get 
      :: "'capacity graph \<Rightarrow> edge \<Rightarrow> 'capacity nres" 
      where "cf_get cf e \<equiv> ASSERT (valid_edge e) \<then> RETURN (cf e)"  
    definition cf_set 
      :: "'capacity graph \<Rightarrow> edge \<Rightarrow> 'capacity \<Rightarrow> 'capacity graph nres"
      where "cf_set cf e cap \<equiv> ASSERT (valid_edge e) \<then> RETURN (cf(e:=cap))"  

    definition resCap_cf_impl :: "'capacity graph \<Rightarrow> path \<Rightarrow> 'capacity nres" 
    where "resCap_cf_impl cf p \<equiv> 
      case p of
        [] \<Rightarrow> RETURN (0::'capacity)
      | (e#p) \<Rightarrow> do {
          cap \<leftarrow> cf_get cf e;
          ASSERT (distinct p);
          nfoldli 
            p (\<lambda>_. True)
            (\<lambda>e cap. do {
              cape \<leftarrow> cf_get cf e;
              RETURN (min cape cap)
            }) 
            cap
        }"

    lemma (in RGraph) resCap_cf_impl_refine:   
      assumes AUG: "cf.isSimplePath s p t"
      shows "resCap_cf_impl cf p \<le> SPEC (\<lambda>r. r = resCap_cf cf p)"
    proof -
      (* TODO: Can we exploit Min.set_eq_fold *)

      note [simp del] = Min_insert
      note [simp] = Min_insert[symmetric]
      from AUG[THEN cf.isSPath_distinct]
      have "distinct p" .
      moreover from AUG cf.isPath_edgeset have "set p \<subseteq> cf.E"
        by (auto simp: cf.isSimplePath_def)
      hence "set p \<subseteq> Collect valid_edge"  
        using cf.E_ss_VxV by simp
      moreover from AUG have "p\<noteq>[]" by (auto simp: s_not_t) 
        then obtain e p' where "p=e#p'" by (auto simp: neq_Nil_conv)
      ultimately show ?thesis  
        unfolding resCap_cf_impl_def resCap_cf_def cf_get_def
        apply (simp only: list.case)
        apply (refine_vcg nfoldli_rule[where 
            I = "\<lambda>l l' cap. 
              cap = Min (cf`insert e (set l)) 
            \<and> set (l@l') \<subseteq> Collect valid_edge"])
        apply (auto intro!: arg_cong[where f=Min])
        done
    qed    

    definition (in Graph) 
      "augment_edge e cap \<equiv> (c(
                  e := c e - cap, 
        prod.swap e := c (prod.swap e) + cap))"

    (* TODO: This would be much simpler to prove if we had a characterization 
      of simple-path only depending on p. *)    
    lemma (in Graph) augment_cf_inductive:
      fixes e cap
      defines "c' \<equiv> augment_edge e cap"
      assumes P: "isSimplePath s (e#p) t"
      shows "augment_cf (insert e (set p)) cap = Graph.augment_cf c' (set p) cap"
      and "\<exists>s'. Graph.isSimplePath c' s' p t"  
    proof -
      obtain u v where [simp]: "e=(u,v)" by (cases e)

      from isSPath_no_selfloop[OF P] have [simp]: "\<And>u. (u,u)\<notin>set p" "u\<noteq>v" by auto

      from isSPath_nt_parallel[OF P] have [simp]: "(v,u)\<notin>set p" by auto
      from isSPath_distinct[OF P] have [simp]: "(u,v)\<notin>set p" by auto

      show "augment_cf (insert e (set p)) cap = Graph.augment_cf c' (set p) cap"
        apply (rule ext)  
        unfolding Graph.augment_cf_def c'_def Graph.augment_edge_def
        by auto

      have "Graph.isSimplePath c' v p t"  
        unfolding Graph.isSimplePath_def
        apply rule
        apply (rule transfer_path)
        unfolding Graph.E_def
        apply (auto simp: c'_def Graph.augment_edge_def) []
        using P apply (auto simp: isSimplePath_def) []
        using P apply (auto simp: isSimplePath_def) []
        done
      thus "\<exists>s'. Graph.isSimplePath c' s' p t" .. 

    qed    
        
    definition "augment_edge_impl cf e cap \<equiv> do {
      v \<leftarrow> cf_get cf e; cf \<leftarrow> cf_set cf e (v-cap);
      let e = prod.swap e;
      v \<leftarrow> cf_get cf e; cf \<leftarrow> cf_set cf e (v+cap);
      RETURN cf
    }"

    lemma augment_edge_impl_refine: 
      assumes "valid_edge e" "\<forall>u. e\<noteq>(u,u)"
      shows "augment_edge_impl cf e cap 
          \<le> (spec r. r = Graph.augment_edge cf e cap)"
      using assms
      unfolding augment_edge_impl_def Graph.augment_edge_def 
      unfolding cf_get_def cf_set_def
      apply refine_vcg
      apply auto
      done
      
    definition augment_cf_impl 
      :: "'capacity graph \<Rightarrow> path \<Rightarrow> 'capacity \<Rightarrow> 'capacity graph nres" 
      where
      "augment_cf_impl cf p x \<equiv> do {
        (rec\<^sub>T D. \<lambda>
          ([],cf) \<Rightarrow> return cf
        | (e#p,cf) \<Rightarrow> do {
            cf \<leftarrow> augment_edge_impl cf e x;
            D (p,cf)
          }  
        ) (p,cf)
      }"

    text \<open>Deriving the corresponding recursion equations\<close>  
    lemma augment_cf_impl_simps[simp]: 
      "augment_cf_impl cf [] x = return cf"
      "augment_cf_impl cf (e#p) x = do { 
        cf \<leftarrow> augment_edge_impl cf e x; 
        augment_cf_impl cf p x}"
      apply (simp add: augment_cf_impl_def)
      apply (subst RECT_unfold, refine_mono)
      apply simp
      
      apply (simp add: augment_cf_impl_def)
      apply (subst RECT_unfold, refine_mono)
      apply simp
      done      

    lemma augment_cf_impl_aux:    
      assumes "\<forall>e\<in>set p. valid_edge e"
      assumes "\<exists>s. Graph.isSimplePath cf s p t"
      shows "augment_cf_impl cf p x \<le> RETURN (Graph.augment_cf cf (set p) x)"
      using assms
      apply (induction p arbitrary: cf)
      apply (simp add: Graph.augment_cf_empty)

      apply clarsimp
      apply (subst Graph.augment_cf_inductive, assumption)

      apply (refine_vcg augment_edge_impl_refine[THEN order_trans])
      apply simp
      apply simp
      apply (auto dest: Graph.isSPath_no_selfloop) []
      apply (rule order_trans, rprems)
        apply (drule Graph.augment_cf_inductive(2)[where cap=x]; simp)
        apply simp
      done  
      
    lemma (in RGraph) augment_cf_impl_refine:     
      assumes "Graph.isSimplePath cf s p t"
      shows "augment_cf_impl cf p x \<le> RETURN (Graph.augment_cf cf (set p) x)"
      apply (rule augment_cf_impl_aux)
      using assms cf.E_ss_VxV apply (auto simp: cf.isSimplePath_def dest!: cf.isPath_edgeset) []
      using assms by blast

      
      
      
      
      
      
      
    subsection \<open>Implementing the Successor Function for BFS\<close>  

    text \<open>We implement the successor function in two steps.
      The first step shows how to obtain the successor function by
      filtering the list of adjacent nodes. This step contains the idea   
      of the implementation. The second step is purely technical, and makes 
      explicit the recursion of the filter function as a recursion combinator
      in the monad. This is required for the Sepref tool.
      \<close>

    text \<open>Note: We use @{term filter_rev} here, as it is tail-recursive, 
      and we are not interested in the order of successors.\<close>
    definition "rg_succ am cf u \<equiv>  
      filter_rev (\<lambda>v. cf (u,v) > 0) (am u)"
  
    lemma (in RGraph) rg_succ_ref1: "\<lbrakk>is_adj_map am\<rbrakk> 
      \<Longrightarrow> (rg_succ am cf u, Graph.E cf``{u}) \<in> \<langle>Id\<rangle>list_set_rel"
      unfolding Graph.E_def
      apply (clarsimp simp: list_set_rel_def br_def rg_succ_def filter_rev_alt; 
        intro conjI)
      using cfE_ss_invE resE_nonNegative 
      apply (auto 
        simp: is_adj_map_def less_le Graph.E_def 
        simp del: cf.zero_cap_simp zero_cap_simp) []
      apply (auto simp: is_adj_map_def) []
      done

    definition ps_get_op :: "_ \<Rightarrow> node \<Rightarrow> node list nres" 
      where "ps_get_op am u \<equiv> assert (u\<in>V) \<then> return (am u)"

    definition monadic_filter_rev_aux 
      :: "'a list \<Rightarrow> ('a \<Rightarrow> bool nres) \<Rightarrow> 'a list \<Rightarrow> 'a list nres"
    where
      "monadic_filter_rev_aux a P l \<equiv> (rec\<^sub>T D. (\<lambda>(l,a). case l of
        [] \<Rightarrow> return a 
      | (v#l) \<Rightarrow> do {
          c \<leftarrow> P v;
          let a = (if c then v#a else a);
          D (l,a)
        }
      )) (l,a)"

    lemma monadic_filter_rev_aux_rule:
      assumes "\<And>x. x\<in>set l \<Longrightarrow> P x \<le> SPEC (\<lambda>r. r=Q x)"
      shows "monadic_filter_rev_aux a P l \<le> SPEC (\<lambda>r. r=filter_rev_aux a Q l)"
      using assms
      apply (induction l arbitrary: a)

      apply (unfold monadic_filter_rev_aux_def) []
      apply (subst RECT_unfold, refine_mono)
      apply (fold monadic_filter_rev_aux_def) []
      apply simp

      apply (unfold monadic_filter_rev_aux_def) []
      apply (subst RECT_unfold, refine_mono)
      apply (fold monadic_filter_rev_aux_def) []
      apply (auto simp: pw_le_iff refine_pw_simps)
      done

    definition "monadic_filter_rev = monadic_filter_rev_aux []"

    lemma monadic_filter_rev_rule:
      assumes "\<And>x. x\<in>set l \<Longrightarrow> P x \<le> (spec r. r=Q x)"
      shows "monadic_filter_rev P l \<le> (spec r. r=filter_rev Q l)"
      using monadic_filter_rev_aux_rule[where a="[]"] assms
      by (auto simp: monadic_filter_rev_def filter_rev_def)

    definition "rg_succ2 am cf u \<equiv> do {
      l \<leftarrow> ps_get_op am u;
      monadic_filter_rev (\<lambda>v. do {
        x \<leftarrow> cf_get cf (u,v);
        return (x>0)
      }) l
    }"

    lemma (in RGraph) rg_succ_ref2: 
      assumes PS: "is_adj_map am" and V: "u\<in>V"
      shows "rg_succ2 am cf u \<le> return (rg_succ am cf u)"
    proof -
      have "\<forall>v\<in>set (am u). valid_edge (u,v)"
        using PS V
        by (auto simp: is_adj_map_def Graph.V_def)
      
      thus ?thesis  
        unfolding rg_succ2_def rg_succ_def ps_get_op_def cf_get_def
        apply (refine_vcg monadic_filter_rev_rule[
            where Q="(\<lambda>v. 0 < cf (u, v))", THEN order_trans])
        by (vc_solve simp: V)
    qed    

    lemma (in RGraph) rg_succ_ref:
      assumes A: "is_adj_map am"
      assumes B: "u\<in>V"
      shows "rg_succ2 am cf u \<le> SPEC (\<lambda>l. (l,cf.E``{u}) \<in> \<langle>Id\<rangle>list_set_rel)"
      using rg_succ_ref1[OF A, of u] rg_succ_ref2[OF A B]
      by (auto simp: pw_le_iff refine_pw_simps)


    subsection \<open>Adding Tabulation of Input\<close>  
    text \<open>
      Next, we add functions that will be refined to tabulate the input of 
      the algorithm, i.e., the network's capacity matrix and adjacency map,
      into efficient representations. 
      The capacity matrix is tabulated to give the initial residual graph,
      and the adjacency map is tabulated for faster access.

      Note, on the abstract level, the tabulation functions are just identity,
      and merely serve as marker constants for implementation.
      \<close>
    definition init_cf :: "'capacity graph nres" 
      \<comment> \<open>Initialization of residual graph from network\<close>
      where "init_cf \<equiv> RETURN c"
    definition init_ps :: "(node \<Rightarrow> node list) \<Rightarrow> _" 
      \<comment> \<open>Initialization of adjacency map\<close>
      where "init_ps am \<equiv> ASSERT (is_adj_map am) \<then> RETURN am"

    definition compute_rflow :: "'capacity graph \<Rightarrow> 'capacity flow nres" 
      \<comment> \<open>Extraction of result flow from residual graph\<close>
      where
      "compute_rflow cf \<equiv> ASSERT (RGraph c s t cf) \<then> RETURN (flow_of_cf cf)"

    definition "bfs2_op am cf \<equiv> Graph.bfs2 cf (rg_succ2 am cf) s t"
      
      
      
      
      
      
      
      
      
  end










end