(**
  To be viewed in Isabelle/jedit IDE. In a normal text editor, this will just be UNREADABLE!

  http://isabelle.in.tum.de

*)
theory Edka_Refine_Demo
  (*
    We import some background theory
  *)
  imports Edka_Refine_Demo_Background
begin


context Network begin

(*
  We add an assertion to the abstract algorithm, stating that
  the selected path is actually an augmenting path.
  
  This is straightforward when proving the abstract algorithm correct.
*)

definition edmonds_karp where "edmonds_karp \<equiv> do {
  let f = (\<lambda>_. 0);
  (f,_) \<leftarrow> while\<^sub>T
    (\<lambda>(f,brk). \<not>brk) 
    (\<lambda>(f,_). do {
      p \<leftarrow> select p. Graph.isShortestPath (cf_of f) s p t;
      case p of 
        None \<Rightarrow> return (f,True)
      | Some p \<Rightarrow> do {
          assert (NPreflow.isAugmentingPath c s t f p); \<comment> \<open>Added assertion\<close>
          let f = NFlow.augment_with_path c f p;
          return (f, False)
        }  
    })
    (f,False);
  return f 
}"

(* Actually, the correctness proof does not change at all! *)

abbreviation edka_invar where "edka_invar \<equiv> \<lambda>(f,brk). 
        NFlow c s t f 
      \<and> (brk \<longrightarrow> (\<forall>p. \<not>NPreflow.isAugmentingPath c s t f p))
    "  

definition "edka_wf_rel \<equiv> inv_image 
  (less_than_bool <*lex*> measure (\<lambda>cf. ek_analysis_defs.ekMeasure cf s t))
  (\<lambda>(f,brk). (\<not>brk,residualGraph c f))"

lemma edka_wf_rel_wf[simp, intro!]: "wf edka_wf_rel"
  unfolding edka_wf_rel_def by auto

lemma edmonds_karp_correct: "edmonds_karp \<le> (spec f. isMaxFlow f)"
  unfolding edmonds_karp_def
  apply (refine_vcg WHILET_rule[where I=fofu_invar and R=edka_wf_rel])
  apply (auto simp: 
    zero_flow 
    NFlow.augment_pres_nflow 
    NFlow.augmenting_path_not_empty
    NFlow.noAugPath_iff_maxFlow[symmetric]
    NFlow.augment_with_path_def
    edka_wf_rel_def NFlow.shortest_path_decr_ek_measure
    NFlow.shortest_is_augmenting
    dest: NFlow.augmenting_path_imp_shortest
  )
  done
  
  

(** At this point, we have (imported from background theory) *)

term cfi_rel
thm cfi_rel_def
  
thm zero_flow_cfi_refine   \<comment> \<open>To refine initialization\<close>
thm augment_cfi_refine     \<comment> \<open>To refine augmentation\<close>
thm cf_of_cfi_rel          \<comment> \<open>Shortest path computed on residual graph\<close>
thm flow_of_cfi_rel        \<comment> \<open>Compute flow from residual graph\<close>
  
  
definition edmonds_karp2 where "edmonds_karp2 \<equiv> do {
  let cf = c;
  (cf,_) \<leftarrow> while\<^sub>T
    (\<lambda>(cf,brk). \<not>brk) 
    (\<lambda>(cf,_). do {
      p \<leftarrow> select p. Graph.isShortestPath cf s p t;
      case p of 
        None \<Rightarrow> return (cf,True)
      | Some p \<Rightarrow> do {
          let cf = Graph.augment_cf cf (set p) (resCap_cf cf p);
          return (cf, False)
        }  
    })
    (cf,False);
  return (flow_of_cf cf)
}"
  

(* Do refinement *)  

text \<open>Automatic proof\<close>

lemma "edmonds_karp2 \<le> \<Down>Id edmonds_karp"
  unfolding edmonds_karp2_def edmonds_karp_def
  \<comment> \<open>Register \<open>cfi_rel\<close> for relation guessing heuristics\<close>
  supply RELATESI[of cfi_rel, refine_dref_RELATES]
  \<comment> \<open>Invoke VCG\<close>
  apply refine_vcg
  \<comment> \<open>Guess Relations\<close>
  apply refine_dref_type
  \<comment> \<open>Prove VCs\<close>
  apply (auto simp: zero_flow_cfi_refine cf_of_cfi_rel flow_of_cfi_rel augment_cfi_refine)
  done




(* Manually, to illustrate what happens behind the scenes *)
lemma edka2_refine: "edmonds_karp2 \<le> \<Down>Id edmonds_karp"
  unfolding edmonds_karp2_def edmonds_karp_def
  apply (rule refine) (* let *)
  apply (rule refine) (* bind *)
  apply (rule refine) (* while *)
  apply (rule introR[where R=\<open>cfi_rel \<times>\<^sub>r Id\<close>]) (* Instantiate relation for while-state *)
  apply (simp add: zero_flow_cfi_refine) (* Refinement: initial state *)
  apply auto [] (* Refinement: while-condition *)
  apply (simp only: split: prod.split; intro allI impI) (* Product (Gets special treatment in VCG) *)
  apply (rule refine) (* bind *)
  apply (rule refine) (* select *) thm SELECT_refine
  apply (rule RELATESI[where R=Id]) (* Instantiate relation for path *)
  apply (auto simp: cf_of_cfi_rel) [] (* Refinement: is_shortest_path *)
  apply (auto simp: cf_of_cfi_rel) [] (* Refinement: is_shortest_path *)
  apply (rule refine) (* option-case *)
  apply assumption    (* Refinement: Operand of option-case *)
  apply (rule refine) (* return *)
  apply auto []       (* Refinement: return with break *)
  apply (rule refine) (*** Congruence for assertion  *)
  apply (rule refine) (* let *)
  apply (rule refine) (* return *)
  apply (auto simp: augment_cfi_refine) [] (* Refinement: augmentation *)
  apply (simp only: split: prod.split; intro allI impI) (* prod-case *)
  apply (rule refine) (* return *)
  apply (auto simp: flow_of_cfi_rel) [] (* Refinement: flow_of_cf *)
  done
  
  


end
end
