theory Edka_Refine_Demo_Background
  imports 
    EdmondsKarp_Maxflow.FordFulkerson_Algo 
    EdmondsKarp_Maxflow.EdmondsKarp_Termination_Abstract
begin

(* Shut up some confusing warnings \<dots> *)
no_notation Heap_Monad.return ("return")

context Network begin

  (******** Some background theory on flows **)
  text \<open>We define the relation between residual graphs and flows\<close>
  definition "cfi_rel \<equiv> br flow_of_cf (RGraph c s t)"
  
  
  text \<open>\<^const>\<open>cfi_rel\<close> can also be characterized the other way round, i.e., 
    mapping flows to residual graphs:\<close>
  lemma cfi_rel_alt: "cfi_rel = {(cf,f). cf = residualGraph c f \<and> NFlow c s t f}"
    unfolding cfi_rel_def br_def
    by (auto 
        simp: NFlow.is_RGraph RGraph.is_NFlow 
        simp: RPreGraph.rg_fo_inv[OF RGraph.this_loc_rpg]
        simp: NPreflow.fo_rg_inv[OF NFlow.axioms(1)])
    
  lemma cf_of_cfi_rel: "(cf,f)\<in>cfi_rel \<Longrightarrow> cf_of f = cf"
    unfolding cfi_rel_alt
    by (auto simp: in_br_conv)
  
  lemma flow_of_cfi_rel: "(cf,f)\<in>cfi_rel \<Longrightarrow> flow_of_cf cf = f"
    unfolding cfi_rel_def
    by (auto simp: in_br_conv)
    
  
  
  lemma residualGraph_zero_flow: "residualGraph c (\<lambda>_. 0) = c" 
    unfolding residualGraph_def by (auto intro!: ext)
    
  lemma flow_of_c: "flow_of_cf c = (\<lambda>_. 0)"
    by (auto simp add: flow_of_cf_def[abs_def])
  
  lemma zero_flow_cfi_refine: "(c,\<lambda>_. 0) \<in> cfi_rel"  
  proof -
    have "NFlow c s t (\<lambda>_. 0)" by (rule zero_flow)
    hence "RGraph c s t c" 
      by (auto dest: NFlow.is_RGraph simp: residualGraph_zero_flow) 
    then show ?thesis
      by (auto simp: cfi_rel_def in_br_conv flow_of_c)
  qed  
  
  
  definition "resCap_cf cf p \<equiv> Min {cf e | e. e\<in>set p}"
  lemma (in NFlow) resCap_cf_refine: "resCap_cf cf p = resCap p"
    unfolding resCap_cf_def resCap_def ..
    
  lemma (in NFlow) augment_cfi_refine_aux: (* For snippet *)
    assumes AUG: "isAugmentingPath p"
    shows "residualGraph c (augment (augmentingFlow p)) (u,v) = (
      if (u,v)\<in>set p then (residualGraph c f (u,v) - resCap p)
      else if (v,u)\<in>set p then (residualGraph c f (u,v) + resCap p)
      else residualGraph c f (u,v))"
    using augment_alt[OF AUG] by (auto simp: Graph.augment_cf_def)

        
        
  lemma augment_cfi_refine:
    assumes R: "(cf,f)\<in>cfi_rel"
    assumes AUG: "NPreflow.isAugmentingPath c s t f p"
    shows "(Graph.augment_cf cf (set p) (resCap_cf cf p), 
        NFlow.augment_with_path c f p) \<in> cfi_rel"
  proof -    
    from R have [simp]: "cf = residualGraph c f" and "NFlow c s t f"
      by (auto simp: cfi_rel_alt br_def)
    then interpret f: NFlow c s t f by simp
    
    show ?thesis 
      unfolding f.augment_with_path_def
    proof (simp add: cfi_rel_alt; safe del: ext intro!: ext)
      fix u v
      show "Graph.augment_cf f.cf (set p) (resCap_cf f.cf p) (u,v) 
            = residualGraph c (f.augment (f.augmentingFlow p)) (u,v)"
        unfolding f.augment_cfi_refine_aux[OF AUG]
        unfolding f.cf.augment_cf_def
        by (auto simp: f.resCap_cf_refine)
    qed (rule f.augment_pres_nflow[OF AUG])
  qed  
  
(**********)
  
end

end