(**
  To be viewed in Isabelle/jedit IDE. In a normal text editor, this will just be UNREADABLE!

  http://isabelle.in.tum.de

*)
theory Workset_Demo
imports Refine_Imperative_HOL.IICF
begin
  (* TODO: Move! *)
  lemma finite_psupsetI[intro!]: \<open>\<lbrakk>Q\<subset>Q'; Q' \<subseteq> S\<rbrakk> \<Longrightarrow> (Q',Q)\<in>finite_psupset S\<close>
    by (auto simp: finite_psupset_def)
    
  interpretation Refine_Monadic_Syntax .

  section \<open>Abstract Algorithm\<close>
    
  definition workset_algo :: "('v \<times> 'v) set \<Rightarrow> 'v \<Rightarrow> 'v set nres" where 
    "workset_algo E s \<equiv> do {
      let (V,W) = ({},{s});  \<comment> \<open>Initialize visited nodes and workset\<close>
  
      (V,W) \<leftarrow> while\<^sub>T (\<lambda>(V,W). W\<noteq>{}) (\<lambda>(V,W). do { \<comment> \<open>While workset not empty\<close>
        \<comment> \<open>Pick a node \<open>u\<close> from workset\<close>
        u \<leftarrow> RES W; let W = W - {u};

        \<comment> \<open>If \<open>u\<close> already visited, do nothing\<close>  
        if u\<in>V then return (V,W)
        else do {
          let V = V \<union> {u};    \<comment> \<open>Otherwise, add \<open>u\<close> to visited nodes\<close>
          let W = W \<union> E``{u}; \<comment> \<open>And its successors to the workset\<close>
          return (V,W)
        }
      }) (V,W);

      \<comment> \<open>Return visited set after loop has terminated\<close>
      return V 
    }"

  text \<open>We want to show that this algorithm computes the set of reachable nodes\<close>
  lemma workset_algo_correct:
    assumes finite: "finite (E\<^sup>*``{s})"
    shows "workset_algo E s \<le> (spec V. V=E\<^sup>*``{s})"
    unfolding workset_algo_def
    apply (refine_vcg WHILET_rule[where 
            I="\<lambda>(V,W). 
              E``V \<subseteq> V\<union>W    \<comment> \<open>Search frontier\<close>
            \<and> V\<union>W \<subseteq> E\<^sup>*``{s} \<comment> \<open>Contains only reachable nodes\<close>
            \<and> s\<in>W\<union>V"        \<comment> \<open>Contains start node\<close> 
        and R = "finite_psupset (E\<^sup>*``{s}) <*lex*> finite_psubset"])
    apply (vc_solve)
    (* Sledgehammer finds all proofs *)
    
    apply (auto intro: finite) []
    apply auto []
    apply auto []
    using finite_subset local.finite apply blast
    apply auto []
    apply auto []
    apply (auto) []
    using Image_closed_trancl apply blast
    
    (* Manually refined short proof 
      apply (auto intro: finite dest: Image_closed_trancl dest: finite_subset) *)
    done
      
  
    
section \<open>Refinement to Successor Function\<close>

  definition "E_succ_rel \<equiv> { (succ, {(u,v). v\<in>set (succ u)}) | succ. True}"

  lemma add_succ_refine: \<open>(succ,E)\<in>E_succ_rel \<Longrightarrow> fold insert (succ u) W = W \<union> E``{u} \<close>
    by (auto simp: union_set_fold[symmetric] E_succ_rel_def)
      
  definition "workset_algo2 succ s \<equiv> do {
    let (V,W) = ({},{s});  \<comment> \<open>Initialize visited nodes and workset\<close>

    (V,W) \<leftarrow> while\<^sub>T (\<lambda>(V,W). W\<noteq>{}) (\<lambda>(V,W). do { \<comment> \<open>While workset not empty\<close>
      \<comment> \<open>Pick a node \<open>u\<close> from workset\<close>
      u \<leftarrow> RES W; let W = W - {u};

      \<comment> \<open>If \<open>u\<close> already visited, do nothing\<close>  
      if u\<in>V then return (V,W)
      else do {
        let V = V \<union> {u};    \<comment> \<open>Otherwise, add \<open>u\<close> to visited nodes\<close>
        let W = fold insert (succ u) W; \<comment> \<open>And its successors to the workset\<close>
        return (V,W)
      }
    }) (V,W);

    \<comment> \<open>Return visited set after loop has terminated\<close>
    return V 
  }"

  lemma workset_algo2_refine: "(succ,E)\<in>E_succ_rel \<Longrightarrow> workset_algo2 succ s \<le>\<Down>Id (workset_algo E s)"
    unfolding workset_algo2_def workset_algo_def
    apply refine_vcg
    apply refine_dref_type
    apply (auto simp: add_succ_refine)
    done
    

end
