(**
  To be viewed in Isabelle/jedit IDE. In a normal text editor, this will just be UNREADABLE!

  http://isabelle.in.tum.de

*)
theory Workset_Demo_Impl
imports Refine_Imperative_HOL.IICF
begin

  (*
    Abstractly, we use a multiset, b/c we can easily refine it to a list.
    (We can't for a set)
  
  *)

  (* TODO: Move! *)
  lemma finite_psupsetI[intro!]: \<open>\<lbrakk>Q\<subset>Q'; Q' \<subseteq> S\<rbrakk> \<Longrightarrow> (Q',Q)\<in>finite_psupset S\<close>
    by (auto simp: finite_psupset_def)

  instance integer :: heap
    apply standard 
    apply (rule exI[where x="int_encode o int_of_integer"])
    by (metis (mono_tags, hide_lams) comp_apply inj_def int_encode_inverse integer_eq_iff)
  
  lemma pure_id_refl[simp]: "pure Id a a = emp"
    by (auto simp: pure_def)

  lemma val_id_refl[simp]: "hn_val Id a a = emp"  
    by (auto simp: pure_def hn_ctxt_def)
      
    

    
            
  interpretation Refine_Monadic_Syntax .

  section \<open>Abstract Algorithm\<close>
    
  definition workset_algo :: "('v \<times> 'v) set \<Rightarrow> 'v \<Rightarrow> 'v set nres" where 
    "workset_algo E s \<equiv> do {
      let (V,W) = ({},{#s#});  \<comment> \<open>Initialize visited nodes and workset\<close>
  
      (V,W) \<leftarrow> while\<^sub>T (\<lambda>(V,W). W\<noteq>{#}) (\<lambda>(V,W). do { \<comment> \<open>While workset not empty\<close>
        \<comment> \<open>Pick a node \<open>u\<close> from workset\<close>
        u \<leftarrow> RES (set_mset W); let W = W - {#u#};

        \<comment> \<open>If \<open>u\<close> already visited, do nothing\<close>  
        if u\<in>V then return (V,W)
        else do {
          let V = V \<union> {u};    \<comment> \<open>Otherwise, add \<open>u\<close> to visited nodes\<close>
          W \<leftarrow> spec W'. set_mset W' = set_mset W \<union> (E``{u}); \<comment> \<open>And its successors to the workset\<close>
          return (V,W)
        }
      }) (V,W);

      \<comment> \<open>Return visited set after loop has terminated\<close>
      return V 
    }"

  text \<open>We show that this algorithm computes the set of reachable nodes\<close>

  (*
    The proof is slightly more complicated, due to the use of multisets.
  *)
  lemma workset_algo_correct:
    assumes finite: "finite (E\<^sup>*``{s})"
    shows "workset_algo E s \<le> (spec V. V=E\<^sup>*``{s})"
  proof -

    have aux:
       \<open>x \<notin># A - {#y#} \<longleftrightarrow> (x=y \<and> count A x = 1) \<or> x\<notin>#A\<close>for x A y
      by (smt One_nat_def add_cancel_right_left add_diff_cancel_right' count_add_mset count_diff count_eq_zero_iff count_greater_eq_Suc_zero_iff diff_zero in_diff_count in_diff_countE insert_DiffM less_imp_le)
  
    have aux2: \<open>x \<in># A - {#y#} \<longleftrightarrow> \<not>((x=y \<and> count A x = 1) \<or> x\<notin>#A) \<close> for x A y
      by (meson aux)
    
    have aux3: \<open>\<lbrakk>finite (E\<^sup>*``{s}); x\<in>X; X\<subseteq>E\<^sup>*``{s}\<rbrakk> \<Longrightarrow> finite (E``{x})\<close> for x X
      by (meson Image_singleton_iff finite_Image_subset finite_reachable_advance r_le_rtrancl subset_iff)
    
    show ?thesis
      unfolding workset_algo_def
      apply (refine_vcg WHILET_rule[where 
              I="\<lambda>(V,W). 
                E``V \<subseteq> V\<union>set_mset W    \<comment> \<open>Search frontier\<close>
              \<and> V\<union>set_mset W \<subseteq> E\<^sup>*``{s} \<comment> \<open>Contains only reachable nodes\<close>
              \<and> s\<in>set_mset W\<union>V"        \<comment> \<open>Contains start node\<close> 
          and R = "(finite_psupset (E\<^sup>*``{s})) <*lex*> measure size"])
      apply (vc_solve)
      (* Finding proofs required some work *)
      apply (auto 0 3 simp: size_Diff1_less aux aux2 aux3[OF finite] intro: finite dest: Image_closed_trancl)
      done
  qed      
  
  
  
    
section \<open>Refinement to Successor Function\<close>
  
  (*
    Implement the state space by a function from a state to a list of successor states
  *)

  definition "E_succ_rel \<equiv> { (succ, {(u,v). v\<in>set (succ u)}) | succ. True}"

  (* Refine the add-successors operation *)
  lemma add_mset_fold: "fold add_mset l m = m + mset l"
    apply (induction l arbitrary: m) by auto
    
  lemma add_succ_refine: "(succ,E)\<in>E_succ_rel \<Longrightarrow> set_mset (fold add_mset (succ u) W) = set_mset W \<union> E``{u}"
    by (auto simp: E_succ_rel_def add_mset_fold)

  (* The refined algorithm. We additionally replace the selection of a node 
      u \<leftarrow> RES (set_mset W); let W = W - {#u#};
    by the constant 
      op_mset_pick
    which is the same.
    
    This is to illustrate manual structure changing refinements. In practice, we
    would have used op_mset_pick from the beginning.
  *)        
  definition "workset_algo2 succ s \<equiv> do {
    let (V,W) = ({},{#s#});  \<comment> \<open>Initialize visited nodes and workset\<close>

    (V,W) \<leftarrow> while\<^sub>T (\<lambda>(V,W). W\<noteq>{#}) (\<lambda>(V,W). do { \<comment> \<open>While workset not empty\<close>
      \<comment> \<open>Pick a node \<open>u\<close> from workset\<close>
      (u,W) \<leftarrow> mop_mset_pick W;

      \<comment> \<open>If \<open>u\<close> already visited, do nothing\<close>  
      if u\<in>V then return (V,W)
      else do {
        let V = V \<union> {u};    \<comment> \<open>Otherwise, add \<open>u\<close> to visited nodes\<close>
        let W = fold add_mset (succ u) W; \<comment> \<open>And its successors to the workset\<close>
        return (V,W)
      }
    }) (V,W);

    \<comment> \<open>Return visited set after loop has terminated\<close>
    return V 
  }"
  
  lemma manual_refine_pick: "do {u \<leftarrow> RES (set_mset W); let W = W - {#u#}; f u W} = do { (u,W) \<leftarrow> op_mset_pick W; f u W }"
    by (force simp: refine_pw_simps pw_eq_iff)

  lemma workset_algo2_refine: "(succ,E)\<in>E_succ_rel \<Longrightarrow> workset_algo2 succ s \<le>\<Down>Id (workset_algo E s)"
    unfolding workset_algo2_def workset_algo_def
    apply (rewrite manual_refine_pick)
    apply refine_vcg
    apply refine_dref_type
    apply (auto simp: add_succ_refine)
    done
  
section \<open>Refinement to Imperative\<close>    
  

  text \<open>Our translation does not like plain function application, when the head of the function 
    is not a constant, as for the successor function. However, it's easy to work around:\<close>

  definition \<open>op_graph_succ (succ :: 'a \<Rightarrow> 'a list) u \<equiv> succ u\<close>  
  lemma [sepref_import_param]: "(op_graph_succ, op_graph_succ) \<in> Id \<rightarrow> Id \<rightarrow> \<langle>Id\<rangle>list_rel" by simp

  lemma [def_pat_rules]: "(\<union>)$V$(insert$u${}) \<equiv> insert$u$V" by simp

  sepref_definition workset_algo3 is "uncurry workset_algo2" :: \<open>id_assn\<^sup>k *\<^sub>a id_assn\<^sup>k \<rightarrow>\<^sub>a hs.assn id_assn\<close>
    unfolding workset_algo2_def
    \<comment> \<open>Determine data structures, by tagging constructors\<close>
    apply (rewrite at "(\<hole>,_)" hs.fold_custom_empty)
    apply (rewrite in "(_,\<hole>)" lms_fold_custom_empty)
    \<comment> \<open>Introduce constant for get-successors operation\<close>
    apply (rewrite in "fold _ \<hole>" op_graph_succ_def[symmetric])
    by sepref

  export_code workset_algo3 in SML_imp module_name Workset
    (* generated code, do not read ;) *)

  (* However, the generated code can be easily interfaced with: *)

  (* To get rid of a type-class dictionary construction, we specialize *)      
  definition workset_algo3_spec :: "(integer \<Rightarrow> integer list) \<Rightarrow> integer \<Rightarrow> integer list Heap" where
    "workset_algo3_spec succ s \<equiv> workset_algo3 succ s \<bind> hs.to_list"

  ML_val \<open>
    fun succ 0 = [1,2]
      | succ 1 = [3]
      | succ 2 = [0]
      | succ _ = []
      
   
    val reachable = 
      @{code workset_algo3_spec} succ 2 ()
  \<close>  
    
    

  (* Finally, we have to combine the refinement statements by transitivity,
    to reach a hoare-triple.
    
    Unfortunately, we need some boilerplate to convert the prior refinements
    to parametricity notation.
    
    But then, the proof is straightforward.
  *)

  thm workset_algo_correct
  lemma workset_algo_correct': "(uncurry workset_algo, \<lambda>(E, s). spec V. V = E\<^sup>* `` {s}) 
    \<in> [\<lambda>(a, b). finite (a\<^sup>* `` {b})]\<^sub>f Id \<times>\<^sub>r Id \<rightarrow> \<langle>Id\<rangle>nres_rel"
    using workset_algo_correct
    by (force intro!: frefI nres_relI)
  
  thm workset_algo2_refine  
  lemma workset_algo2_refine': "(workset_algo2, workset_algo) \<in> E_succ_rel \<rightarrow> Id \<rightarrow> \<langle>Id\<rangle>nres_rel"  
    using workset_algo2_refine by (auto intro!: nres_relI)
    
  theorem correct: 
    shows "(uncurry workset_algo3, \<lambda>(E, s). spec V. V = E\<^sup>* `` {s})
      \<in> [\<lambda>(a, b). finite (a\<^sup>* `` {b})]\<^sub>a (pure E_succ_rel)\<^sup>k *\<^sub>a id_assn\<^sup>k \<rightarrow> hs.assn id_assn"
    using workset_algo3.refine[FCOMP workset_algo2_refine', FCOMP workset_algo_correct']  
    .

end
